/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *  STRINGS.CPP -- STRING class                        V1.1  *
 *                                                           *
 *  Language:      Borland C++                               *
 *  System:        Gateway 386/25, MS-DOS 5.0                *
 *  Application:   CSE 687                                   *
 *  Author:        Jim Fawcett, CST 2-127, (315)443-4454     *
 *                                                           *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include <stdlib.h>        // ANSI:     exit
#include <ctype.h>         // ANSI:     isspace
#include <string.h>        // ANSI:     strlen, strstr
#include "strings.hpp"     // FAWCETT:  prototypes for strings


//----< string::string(void) >-------------------------------//

//  Processing:  default constructor builds a string with no
//               characters
//  Invocation:  string str;

string::string(void) {

  slen  = 0;
  shead = sptr = '\0';
}

//----< string::string(string &s) >---------------------------//

//  Processing:  copy constructor builds string identical to
//               an existing string
//  Invocation:  string str1 = str2;

string::string(const string &s) {

  slen  = s.slen;
  if (slen < 0) {
    cerr << "dimension error in string(string &s)\n";
    exit(1);
  }
  if (slen ==0)
    sptr = shead = '\0';
  else {
    message("allocating ");
    message(slen+1);
    message(" bytes of heap memory\n");
    if((shead = new char[slen+1])==NULL) {
      cerr << "dynamic memory exhausted\n";
      exit(1);
    }
    strcpy(shead, s.shead);
    int offset = int(s.sptr - s.shead);
    sptr  = shead + offset;
  }
}
/**/
//----< string::string(char *sp) >---------------------------//

//  Processing:  constructs string with characters from a
//               "C" string
//  Invocation:  string str("hello world\n");
//          or:  string str = "hello world\n";

string::string(char *sp) {

  slen = strlen(sp);
  message("allocating ");
  message(slen + 1);
  message(" bytes of heap memory\n");
  if((shead = new char[slen+1])==NULL) {
    cerr << "dynamic memory exhausted\n";
    exit(1);
  }
  strcpy(shead, sp);
  sptr = shead;
}

//----< string::string(int len) >----------------------------//

//  Processing:  constructs string with specified number of
//               blank characters
//  Invocation:  string str(5);

string::string(int len) { // note that len = number of chars
																// not counting null terminator
  slen  = len;
  if (slen < 0) {
    cerr << "dimension error in string(int len)\n";
    exit(1);
  }
  message("allocating ");
  message(slen + 1);
  message(" bytes of heap memory\n");
  if((shead = new char[slen+1])==NULL) {
    cerr << "dynamic memory exhausted\n";
    exit(1);
  }
  for (int i=0; i<slen; i++)
    *(shead+i) = ' ';
  *(shead+slen) = '\0';         // null terminate string
  sptr  = shead;
}

//----< string::~string(void) >------------------------------//

//  Processing:  returns character space to heap
//  Invocation:  implicit

string::~string(void) {

  if(shead) {
    message("deallocating ");
    message(slen+1);
    message(" bytes of heap memory\n");
    delete [] shead;
  }
}
/**/
//----< string& string::operator= (string &s) >--------------//

//  Processing:  assignment operator replaces existing string
//               with a copy of a provide string
//  Invocation:  str2 = str1;

string& string::operator= (const string &str) {

  if (*this == str)
    return *this;
  if (shead) {
    message("deallocating ");
    message(slen+1);
    message(" bytes of heap memory\n");
    delete [] shead;
  }
  slen = str.slen;
  if (slen == 0)
    sptr = shead = '\0';
  else {
    message("allocating ");
    message(slen + 1);
    message(" bytes of heap memory\n");
    if((shead = new char[slen+1])==NULL) {
      cerr << "dynamic memory exhausted\n";
      exit(1);
    }
    strcpy(shead, str.shead);
    sptr  = shead;
  }
  return *this;
}

//----< string& string::operator+=(string &str) >------------//

//  Processing:  self concatenation operator concatenates
//               provided string with itself
//  Invocation:  str2 += str1;

string& string::operator+=(const string &str) {

  slen = slen + str.slen;
  char *newshead;
  message("allocating ");
  message(slen + 1);
  message(" bytes of heap memory\n");
  if((newshead = new char[slen+1])==NULL) {
    cerr << "dynamic memory exhausted\n";
    exit(1);
  }
  strcpy(newshead, shead);
  strcat(newshead, str.shead);
  if(shead) {
    message("deallocating ");
    message(strlen(shead)+1);
    message(" bytes of heap memory\n");
    delete [] shead;
  }
  shead = sptr = newshead;
  return *this;
}
/**/
//----< string operator+ (string &str1, string &str2) >------//

//  Processing:  concatenation operator builds a string formed
//               by concatenating two provided strings
//  Invocation:  str3 = str1 + str2;
//  Note:        can't return reference, e.g., must copy on
//               return since result is built in temporary

string operator+ (const string &str1, const string &str2) {

  int slen = str1.slen + str2.slen;
  char *sp;
  message("allocating ");
  message(slen + 1);
  message(" bytes of heap memory\n");
  if((sp = new char[slen+1])==NULL) {
    cerr << "dynamic memory exhausted\n";
    exit(1);
  }
  strcpy(sp, str1.shead);
  strcat(sp+str1.slen, str2.shead);
  string st(sp);
  return st;
}

//----< char* match(string &str1, string &str2) >------------//

//  Processing:  find position of str2 embedded in str1 if it
//               exists in str1, otherwise return null pointer
//  Invocation:  char *sptr = match(str1, str2);

char* match(string &str1, const string &str2) {

  char *ptr = strstr(str1.shead, str2.shead);
  if (ptr)
    str1.sptr = ptr;
  else
    str1.sptr = str1.shead;
  return ptr;
}
/**/
//----< int select(string &str, int i) >---------------------//

//  Processing:  add i to current index and return char at
//               that location.  Return 0 if outside string
//               char array.
//  Invocation:  int ch = select(str,-1);

int select(string &str, int i) {

  int len = int(str.sptr - str.shead) + i;
  if ((len < 0) || (str.slen < len))
    return '\0';
  str.sptr = str.sptr + i;
  return *str.sptr;
}

//----< string head(string &str) ----------------------------//

//  Processing:  returns a string equal to the string between
//               str.shead and str.sptr not including the later
//  Invocation:  char *sptr = head(str1);

string head(string &str) {

  string null = "";
  int len = int(str.sptr - str.shead);
  if (len < 1)
    return null;
  string s(len);
  char *save = str.sptr;
  for (int i=0; i<len; i++)
    s[i] = str[i];
  str.sptr = save;
  return s;
}

//----< string tail(string &str) ----------------------------//

//  Processing:  returns a string equal to the string starting
//               at str.sptr
//  Invocation:  char *sptr = tail(str1);

string tail(string &str) {

  string null = "";
  int len = length(str) - int(str.sptr - str.shead);
  if (len < 1)
    return null;
  string s(len);
  char *save = str.sptr;
  while((*(s.sptr++) = *(str.sptr++)) != 0)
    ;
  str.sptr = save;
  s.sptr = s.shead;
  return s;
}
/**/
//----< string string::overlay(string &str, int n) >---------//

//  Processing:  overlay str on self at index n, extending
//               length of self if necessary
//  Invocation:  str1.overlay(str2,5);

string string::overlay(const string &str, int n) {

  int len = str.slen + n;
  if (len < slen)
    len = slen;
  string s(len);
  int i;
  for (i=0; i<slen; i++)
    *(s.shead+i) = *(shead+i);
  for (i=0; i<str.slen; i++)
    *(s.shead+n+i) = *(str.shead+i);
  return s;
}

//----< string string::insert(string &str, int n) >----------//

//  Processing:  insert str in self at index n, extending
//               length of self
//  Invocation:  str1.insert(str2,5);

string string::insert(const string &str, int n) {

  int len = slen + str.slen;
  string s(len);
  int i;
  for (i=0; i<n; i++)
    *(s.shead+i) = *(shead+i);
  for (i=0; i<str.slen; i++)
    *(s.shead+n+i) = *(str.shead+i);
  for (i=n; i<slen; i++)
    *(s.shead+str.slen+i) = *(shead+i);
  return s;
}
/**/
//----< char* skipwhite(string &str) >------------------------//

//  Processing:  starting at current position, iterate down
//               string until encounter first nonwhitespace
//               character
//  Invocation:  skipwhite(str);

char* skipwhite(string &str) {

  while(isspace(*(str.sptr)))
    if (int(str.sptr - str.shead) < str.slen)
      str.sptr++;
  return str.sptr;
}

//----< string eatwhite(string &str) >-----------------------//

//  Processing:  starting at string head, iterate until
//               encounter first nonwhitespace character,
//               return tail
//  Invocation:  str2 = eatwhite(str1);

string eatwhite(string &str) {

  string s = skipwhite(str);
  return tail(s);
}
/**/
//----< istream& operator>>(istream &istrm, string &str) >---//

//  Processing:  Read record terminated by newline.  If record
//               length equals zero, continue reading records
//               until a record is found with non-zero length,
//               or reach EOF
//  Invocation:  istrm >> str;

istream& operator>>(istream& istrm, string& str) {

  #define  MAXSIZE 4095
  char* buffer;
  message("allocating ");
  message(MAXSIZE+1);
  message(" bytes of heap memory\n");
  if ((buffer = new char[MAXSIZE+1])==NULL) {
    cerr << "dynamic memory exhausted\n";
    exit(1);
  }
  istrm.getline(buffer,MAXSIZE);
  buffer[MAXSIZE] = '\0';
  if(str.shead) {
    message("deallocating ");
    message(str.slen+1);
    message(" bytes of heap memory\n");
    delete [] str.shead;
  }
  str.slen = strlen(buffer);
  message("allocating ");
  message(str.slen+1);
  message(" bytes of heap memory\n");
  if ((str.shead = str.sptr = new char[str.slen+1]) == NULL) {
    cerr << "dynamic memory exhausted\n";
    exit(1);
  }
  strcpy(str.shead,buffer);
  if (buffer) {
    message("deallocating ");
    message(MAXSIZE+1);
    message(" bytes of memory\n");
    delete [] buffer;
  }
  return istrm;
  #undef MAXSIZE
}

//----< ostream& operator<<(ostream &ostrm, string &str) >---//

//  Processing:  send string to output stream
//  Invocation:  ostrm << str;

ostream& operator<<(ostream &ostrm, string &str) {

  return ostrm << str.shead;
}
